module SmartMail where 

import qualified Data.Map as Map
import Data.Maybe
import Data.List
import Data.Char
import Personne
import Trame
import CompteSmail
import System.Directory

type SmartMail = Map.Map Courriel CompteSmail -- Dictionnaire de comptes smail
first_third (f,_,t) = (f,t) -- Petite fonction utilisée dans les tests

-- | Retourne un SmartMail vide
--
-- >>> nombreCompteSmail emptySmartMail
-- 0
emptySmartMail :: SmartMail 
emptySmartMail = Map.empty


-- | Ajout d'un compteSmail 
--
-- Note: Si un compte existe déjà avec le même courriel alors ne rien faire
--
-- >>> s1 = ajoutCompte csmail2 $ ajoutCompte csmail1 emptySmartMail
-- >>> courrielsComptes s1
-- ["nkambou.roger@smail.ca","tato.ange@smail.ca"]
ajoutCompte :: CompteSmail -> SmartMail-> SmartMail
ajoutCompte compteSmail smartMail = 
    if (Map.member (courriel $ personne compteSmail) smartMail) then 
        smartMail
    else
        Map.insert (courriel $ personne compteSmail) compteSmail smartMail

-- | Ajout de plusieurs comptes Smail 
--
-- >>> courrielsComptes $ ajoutComptes [csmail1, csmail2] emptySmartMail
-- ["nkambou.roger@smail.ca","tato.ange@smail.ca"]
ajoutComptes::  [CompteSmail] -> SmartMail -> SmartMail 
ajoutComptes compteSmails smartMail = foldl (\smail c -> ajoutCompte c smail) smartMail compteSmails

-- | Affiche tous les courriels de comptes Smail
--
-- >>> s1 = ajoutCompte csmail2 $ ajoutCompte csmail1 emptySmartMail
-- >>> courrielsComptes s1
-- ["nkambou.roger@smail.ca","tato.ange@smail.ca"]
-- >>>courrielsComptes emptySmartMail
-- []
courrielsComptes ::SmartMail-> [Courriel]
courrielsComptes smail = map (\smail -> courriel $ personne smail) $ Map.elems smail

-- | Retourne nombre de CompteSmail contenu dans SmartMail
--
-- >>> s1 = ajoutCompte csmail2 $ ajoutCompte csmail1 emptySmartMail
-- >>> nombreCompteSmail s1
-- 2
-- >>> nombreCompteSmail emptySmartMail
-- 0
nombreCompteSmail :: SmartMail -> Int
nombreCompteSmail smail = length $ Map.elems smail

-- | Retoune le comptre Smail associé à un courriel dans SmartMail
-- Note : On suppose que le courriel passé en argumemnt est associé à un compte existant du smartmail
--
-- >>> s1 = ajoutCompte csmail2 $ ajoutCompte csmail1 emptySmartMail
-- >>> obtenirCompte "tato.ange@smail.ca" s1
-- CompteSmail "tato.ange@smail.ca":
-- Recus = [Trame (Entete (Date 2020 2 10) "Bienvenue" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !"],
-- Envois = [],
-- Spams = [],
-- Contacts = [("nkambou.roger@smail.ca",Blanc),("noel.alice@smail.ca",Noir)]
obtenirCompte :: Courriel -> SmartMail -> CompteSmail
obtenirCompte courriel smail  = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) courriel smail 

-- | Déterminer la priorité d'un message non spam
-- Note: on suppose que la Trame en input est un message non spam
--    (1) Important= expediteur est un contact direct (est dans la liste de contacts du recepteur) qui n'est pas bloqué ou est "equipesmartmail@smail.ca" 
--    (2) Normal = expediteur est de niveau 2 (est dans la liste de contacts d'au moins un des contacts du recepteur et n'est pas bloqué). Si il est bloqué par au moins un contact, alors il passe au niveau suivant.
--    (3) Faible = expediteur n'est ni un contact direct, n'est ni de niveau 2, n'est ni l'équipe smartMail.ca
--
-- >>> s1 = ajoutCompte csmail5 $ ajoutCompte csmail4 $ ajoutCompte csmail2 $ ajoutCompte csmail1 emptySmartMail
-- >>> prioriteMessage s1 trameBienvenue2
-- Important
-- >>> prioriteMessage s1 tramet
-- Faible
-- >>> prioriteMessage s1 tramett
-- Normal
prioriteMessage :: SmartMail -> Trame -> Priorite
prioriteMessage smail (Trame (Entete _ _ sender receiver) _)
    | (courriel sender) == "equipesmartmail@smail.ca" = Important
    | (length (filter (== (courriel sender)) white) > 0) &&
      (length (filter (== (courriel sender)) black) == 0) = Important

    | (length (filter (== (courriel sender)) white2) > 0) &&
      (length (filter (== (courriel sender)) black2) == 0) = Normal

    | otherwise = Faible

    where
        recv = obtenirCompte (courriel receiver) smail
        contacts_list = listeContacts $ contacts recv
        white = map (\(x,_)->x) $ filter (\(_,b)->b == Blanc) contacts_list
        black = map (\(x,_)->x) $ filter (\(_,b)->b == Noir) contacts_list

        contact2 = map (\(p,_)->p) $ filter (\(_,b)->b == Noir) contacts_list
        allc2 = foldl (++) [] $ map (\c->contacts $ smail Map.! c) contact2
        white2 = map (\(x,_)->courriel x) $ filter (\(_,b)->b == Blanc) allc2
        black2 = map (\(x,_)->courriel x) $ filter (\(_,b)->b == Noir) allc2



-------------------------------------------------------------------
---------- FILTRES ANTI SPAM ET ANTI HAMEÇONNAGE ------------------ 
-------------------------------------------------------------------
-- | Filtrage de l'enveloppe de la trame
-- Ce filtre s’exécute uniquement sur l'entête de la trame et non sur son contenu.
-- Une trame est détectée comme spam par le filtrageEnveloppe si au moins l'une des conditions suivantes est vraie: 
--      (1) tous les caractères de l'objet sont en majuscules (Explications = "classique_enveloppe")
--      (2) l'objet contient au moins 2 points d'exclamation "!" (Explications = "classique_enveloppe")
--      (3) l'objet contient au moins 2 points d'interrogation "?" (Explications = "classique_enveloppe")
--      (4) l'objet est vide (Explications = "objet vide")
--      (5) l'objet contient le caractère "$" (Explications = "classique_enveloppe")
--      (6) l'emetteur a été bloqué (Etat = Noir) par le destinataire   (Explications = "contact bloque")  
--
-- >>> ssm2 = Map.fromList [("nkambou.roger@smail.ca", CompteSmail (Personne "nkambou.roger@smail.ca" ("roger","nkambou")) [Trame (Entete (Date 2018 10 2) "Bienvenue" pers0 pers1) "Bienvenue dans votre boite smartMail !"] [] [] [] []),("robert.julien@smail.ca",CompteSmail (Personne "robert.julien@smail.ca" ("julien","robert")) [] [] [] [] [(Personne "tato.ange@smail.ca" ("",""),Noir)]),("tato.ange@smail.ca", CompteSmail (Personne "tato.ange@smail.ca" ("ange","tato")) [Trame (Entete (Date 2018 10 2) "Bienvenue" pers0 pers1) "Bienvenue dans votre boite smartMail !"] [] [] [] [])]
-- >>> filtrageEnveloppe trame6 ssm2 
-- (Spam,Trame (Entete (Date 2018 5 9) "Bienvenue" (Personne "tato.ange@smail.ca" ("ange","tato")) (Personne "robert.julien@smail.ca" ("julien","robert"))) "Allo Robert","contact bloque")
-- >>> filtrageEnveloppe trame1 ssm2 
-- (Spam,Trame (Entete (Date 2020 1 18) "AB CD EF" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe")
-- >>> filtrageEnveloppe trame2 ssm2 
-- (Spam,Trame (Entete (Date 2019 12 21) "Bi!en! venue!" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe")
-- >>> filtrageEnveloppe trame3 ssm2 
-- (Spam,Trame (Entete (Date 2020 1 1) "?Bien venue?" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe")
-- >>> filtrageEnveloppe trame4 ssm2 
-- (Spam,Trame (Entete (Date 2018 10 5) "" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","objet vide")
-- >>> filtrageEnveloppe trame5 ssm2 
-- (Spam,Trame (Entete (Date 2017 5 7) "Bienvenue $ " (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe")
-- >>> filtrageEnveloppe trameBienvenue1 ssm2 
-- (NonSpam,Trame (Entete (Date 2020 2 10) "Bienvenue" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","")
-- >>> filtrageEnveloppe trameBienvenue2 ssm2 
-- (NonSpam,Trame (Entete (Date 2020 2 10) "Bienvenue" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "nkambou.roger@smail.ca" ("roger","nkambou"))) "Bienvenue dans votre boite smartMail !","")
filtrageEnveloppe :: Trame -> SmartMail -> (TypeMessage, Trame, Explications) 
filtrageEnveloppe trame sm
    | (length (objet trame) > 0) && (length (filter Data.Char.isLower (objet trame)) == 0) = spamResult
    | (length (filter (== '!') (objet trame))) >= 2 = spamResult
    | (length (filter (== '?') (objet trame))) >= 2 = spamResult
    | (length (objet trame) == 0) = (Spam,  trame, "objet vide")
    | (length (filter (== '$') (objet trame))) > 0 = spamResult
    | blockedBy sm (emetteur trame) (receveur trame) = (Spam,  trame,"contact bloque")
    | otherwise = (NonSpam, trame, "")
    where
        spamResult = (Spam,  trame,"classique_enveloppe")

blockedBy :: SmartMail -> String -> String -> Bool
blockedBy sm email receipient = 
    isBlocked email cs
    where
        cs = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) receipient sm 

isBlocked :: String -> CompteSmail -> Bool
isBlocked email (CompteSmail _ _ _ _ _ contacts) = 
    length (filter (\(Personne contactEmail _, etat) -> etat == Noir && contactEmail == email) contacts) > 0

-- | Filtrage du contenu
-- Les filtres de contenu analysent le contenu des messages et détectent les spams qui ont réussi à passer à travers le filtre d'enveloppe.
-- Le principe consiste généralement à détecter des mots précis ou des mots de formes particulières, par exemple des mots qui contiennent des chiffres ou certains symboles (C|AL|S, -1AGRA, 
-- PR0ZAC, ZYBAN and C3LEBREX cred1ted. Il y a donc deux sous catégories d'heuristiques: la détection des mots clés (heuristiqueMotsCles) et la recherche de mots comportant des caractères spéciaux (heuristiqueCaracteresSpeciaux). 
-- Un poids (proportion de mots clés ou de mots comportant les caractères spéciaux) de l'ordre de 10% suffit pour déclarer le message comme spam.
-- Vous devez définir les 3 filtres suivants:
--      (1) filtreSpecifiqueHams pour les mots-clés suivants: "sexe", "sexy", "viagra", "argent", "drogue", "money","credit", "$","chaud", "nu", "click", "amateur", "pics","videos", "gagner","lotterie","heritage".
--      (2) filtreSpecifiquePub pour les mots-clés "offre", "commande", "click", "videos","gratuit","publicité", "special","voyage"
--      (3) filtreCaracteresSpeciaux pour mots contenant un ou l'autre des caractères suivant '|','0'..'9'.
-- On supposera qu'un message, s'il est un spam, l'est à cause de l'un des filtres exclusivement
-- L'explication donnera les indices sur la nature du spam: "hameconnage, XX% de mots suspects", "publicitaire, XX% de mots suspects" et "classique_contenu, XX% de mots comportant des caracteres etranges" (caractères spéciaux dans les mots). 
-- L'information sur la proportion des mots suspects dans le contenu du message est requis pour les filtres de contenu. Vous devez arrondi le pourcentage. 
-- Remarque: Pour cette question, vous devez écrire des sous-fonctions nécessaires. 
--
-- >>> filtrageContenu trame7
-- (Spam,Trame (Entete (Date 2019 18 10) "un message" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans sexe du viagra chaud nu","hameconnage, 57% de mots suspects.")
-- >>> first_third $ filtrageContenu trame8
-- (Spam,"publicitaire, 56% de mots suspects.")
-- >>> first_third $ filtrageContenu trame9
-- (Spam,"classique_contenu, 67% de mots comportant des caracteres etranges.")
-- >>> first_third $ filtrageContenu trame10
-- (Spam,"classique_contenu, 100% de mots comportant des caracteres etranges.")
-- >>> first_third $ filtrageContenu trame11
-- (Spam,"classique_contenu, 100% de mots comportant des caracteres etranges.")
filtrageContenu :: Trame -> (TypeMessage, Trame, Explications)
filtrageContenu trame
    | filter1Pct > 10 = (Spam, trame, "hameconnage, " ++ (show filter1Pct) ++ "% de mots suspects.")
    | filter2Pct > 10 = (Spam, trame, "publicitaire, " ++ (show filter2Pct) ++ "% de mots suspects.")
    | filter3Pct > 10 = (Spam, trame, "classique_contenu, " ++ (show filter3Pct) ++ "% de mots comportant des caracteres etranges.")
    | otherwise = (NonSpam, trame, "")
    where
        contents = contenu trame
        wordsList = words contents
        nWords = length wordsList
        filter1 = length $ filter filter1Fun wordsList
        filter1Pct = div (100 * filter1) nWords
        filter2 = length $ filter filter2Fun wordsList
        filter2Pct = div (100 * filter2) nWords
        filter3 = length $ filter filter3Fun wordsList
        filter3Pct = div (100 * filter3) nWords

filter1Fun :: String -> Bool
filter1Fun "sexe" = True 
filter1Fun "sexy" = True 
filter1Fun "viagra" = True 
filter1Fun "argent" = True 
filter1Fun "drogue" = True 
filter1Fun "credit" = True 
filter1Fun "$" = True 
filter1Fun "chaud" = True 
filter1Fun "nu" = True 
filter1Fun "click" = True 
filter1Fun "money" = True 
filter1Fun "amateur" = True 
filter1Fun "pics" = True 
filter1Fun "videos" = True 
filter1Fun "gagner" = True 
filter1Fun "lotterie" = True 
filter1Fun "heritage" = True 
filter1Fun _ = False

filter2Fun :: String -> Bool
filter2Fun "offer" = True 
filter2Fun "order" = True 
filter2Fun "click" = True 
filter2Fun "videos" = True 
filter2Fun "free" = True 
filter2Fun "advertising" = True 
filter2Fun "special" = True 
filter2Fun "travel" = True
filter2Fun _ = False

filter3Fun :: String -> Bool
filter3Fun word = length (filter (\c -> Data.Char.isDigit(c) || c=='|') word) > 0

-- | Envoyer message
-- Remarque: Le message doit être enregistré dans la boîte d'envoi de l'emetteur mais doit également être enregistré dans la boîte de reception du destinataire (si le message n'est pas un spam c-à-d que filtreEnvelope ou filtreContenu ne retourne pas "Spam") ou dans la boîte de spams le cas échéant. Si le destinataire n'a pas de compte "SmartMail", ne rien faire et ne garder le message que dans la boite de l'émétteur 
--
-- >>> s1 = ajoutCompte csmail2 $ ajoutCompte csmail1 $ ajoutCompte csmail0 emptySmartMail
-- >>> s2 = envoyerMessage  s1 ("equipesmartmail@smail.ca", "tato.ange@smail.ca", "Bi!en! venue!", "Bienvenue dans votre boite smartMail !")  
-- >>> length (envoi $ obtenirCompte "equipesmartmail@smail.ca" s2)
-- 3
-- >>> length (reception $ obtenirCompte "tato.ange@smail.ca" s2)
-- 1
-- >>> s3 = ajoutCompte csmail3 s2
-- >>> s4 = envoyerMessage s3 ("equipesmartmail@smail.ca", "tato.ange@smail.ca", "?Bien venue?", "Bienvenue dans votre boite smartMail !")
-- >>> length (reception $ obtenirCompte "tato.ange@smail.ca" s4)
-- 1
-- >>> length (envoi $ obtenirCompte "tato.ange@smail.ca" s4)
-- 0

envoyerMessage :: SmartMail -> Message -> SmartMail
envoyerMessage sm (sender, receipient, subject, content) = 
    sm4
    where
        senderData = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) sender sm 
        CompteSmail senderPerson senderInbox senderSent senderSpam senderPref senderContact = senderData
        recipientData = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) receipient sm 
        CompteSmail recvPerson recvInbox recvSent recvSpam recvPref recvContact = recipientData
        (y,m,d) = dateAuj 
        trame = Trame (Entete (Date y m d) subject senderPerson recvPerson) content

        sm2 = Map.delete receipient $ Map.delete sender sm
        senderDataWithInbox = CompteSmail senderPerson senderInbox (trame:senderSent) senderSpam senderPref senderContact
        sm3 = ajoutCompte senderDataWithInbox sm2

        (isSpam1, _, exp1) = filtrageEnveloppe trame sm
        (isSpam2, _, exp2) = filtrageContenu trame
        exp = exp1 ++ exp2
        isSpam = isSpam1 == Spam || isSpam2 == Spam

        recvDataNew = if isSpam then
                        CompteSmail recvPerson recvInbox recvSent ((trame, exp):recvSpam) recvPref recvContact
                      else
                        CompteSmail recvPerson (trame:recvInbox) recvSent recvSpam recvPref recvContact
        sm4 = if recvPerson == (Personne "" ("","")) then -- receipient does not have an account
                sm4
              else
                ajoutCompte recvDataNew sm3

sendTrame :: SmartMail -> Trame -> SmartMail
sendTrame sm trame = 
    sm4
    where
        senderData = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) (emetteur trame) sm 
        CompteSmail senderPerson senderInbox senderSent senderSpam senderPref senderContact = senderData
        recipientData = Map.findWithDefault (CompteSmail (Personne "" ("","")) [] [] [] [] []) (receveur trame) sm 
        CompteSmail recvPerson recvInbox recvSent recvSpam recvPref recvContact = recipientData

        sm2 = Map.delete (receveur trame) $ Map.delete (emetteur trame) sm
        senderDataWithInbox = CompteSmail senderPerson senderInbox (trame:senderSent) senderSpam senderPref senderContact
        sm3 = ajoutCompte senderDataWithInbox sm2

        (isSpam1, _, exp1) = filtrageEnveloppe trame sm
        (isSpam2, _, exp2) = filtrageContenu trame
        exp = exp1 ++ exp2
        isSpam = isSpam1 == Spam || isSpam2 == Spam

        recvDataNew = if isSpam then
                        CompteSmail recvPerson recvInbox recvSent ((trame, exp):recvSpam) recvPref recvContact
                      else
                        CompteSmail recvPerson (trame:recvInbox) recvSent recvSpam recvPref recvContact
        sm4 = if recvPerson == (Personne "" ("","")) then -- receipient does not have an account
                sm4
              else
                ajoutCompte recvDataNew sm3

-------------------------------------------------------------------
--------------------   STATISTIQUES ET AFFICHAGES   ---------------   
-------------------------------------------------------------------

-- | Donne le nombre de spams recus dans le systeme de messagerie au complet
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> nbTotalSpams s1
-- 0
-- >>> nbTotalSpams s
-- 8
nbTotalSpams :: SmartMail -> Int
nbTotalSpams sm =
    length $ foldl (++) [] $ map (\(k,a)->spams a) (Map.toList sm)

-- | Retourne l'ensemble de tous les spams du système dans une même liste
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> tousLesSpams s1
-- []
-- >>> tousLesSpams s
-- [(Trame (Entete (Date 2020 10 10) "un message" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "0de 3Bienvenue dans 1 et 9 | pour tp1et2","classique_contenu, 67% de mots comportant des caracteres etranges."),(Trame (Entete (Date 2020 1 15) "un message" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "offre de Bienvenue dans  publicit\233 gratuit pour voyage special","publicitaire, 56% de mots suspects."),(Trame (Entete (Date 2019 18 10) "un message" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans sexe du viagra chaud nu","hameconnage, 57% de mots suspects."),(Trame (Entete (Date 2017 5 7) "Bienvenue $ " (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe"),(Trame (Entete (Date 2018 10 5) "" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","objet vide"),(Trame (Entete (Date 2019 12 21) "Bi!en! venue!" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe"),(Trame (Entete (Date 2019 12 21) "Bi!en! venue!" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe"),(Trame (Entete (Date 2020 1 18) "AB CD EF" (Personne "equipesmartmail@smail.ca" ("equipe","smail")) (Personne "tato.ange@smail.ca" ("ange","tato"))) "Bienvenue dans votre boite smartMail !","classique_enveloppe")]
tousLesSpams:: SmartMail -> [(Trame, Explications)]
tousLesSpams sm = foldl (++) [] $ map (\(k,a)->spams a) (Map.toList sm)

-- | Retourne la liste qui associe à chaque inscrit (son courriel seulement), le nombre de spams recus. Résultat par ordre alphabétique de courriels
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> statSpamsRecus s1
-- [("bourassa.alex@smail.ca",0),("equipesmartmail@smail.ca",0),("nkambou.roger@smail.ca",0),("noel.alice@smail.ca",0),("robert.julien@smail.ca",0),("tato.ange@smail.ca",0)]
-- >>> statSpamsRecus s
-- [("bourassa.alex@smail.ca",0),("equipesmartmail@smail.ca",0),("nkambou.roger@smail.ca",0),("noel.alice@smail.ca",0),("robert.julien@smail.ca",0),("tato.ange@smail.ca",8)]
statSpamsRecus :: SmartMail -> [(String, Int)]
statSpamsRecus sm = sort $ map (\(email,cs)->(email,length $ spams cs)) (Map.toList sm)

-- | Produit une liste qui associe à chaque inscrit (son courriel seulement), le nombre de spams produit
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> statSpamsEnvoyes s1
-- []
-- >>> statSpamsEnvoyes s
-- [("equipesmartmail@smail.ca",8)]
-- statSpamsEnvoyes :: SmartMail -> [(Courriel, Int)]
statSpamsEnvoyes sm = 
    foldl (
        \m s->
            if (Map.member s m) then
                Map.insert s (( m Map.! s) + 1) $ Map.delete s m
            else
                Map.insert s 1 m 
        ) Map.empty allSpammers
    where
        allSpammers = map (\(trame,_)->emetteur trame) $ foldl (++) [] $ map (\(_,cs)-> spams cs) $ Map.toList sm



-- | Envoyer plusieurs trames en même temps
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> length $ spams $ obtenirCompte "tato.ange@smail.ca"  s
-- 8
envoyerMessage_Plusieurstrames :: SmartMail -> [Trame] -> SmartMail
envoyerMessage_Plusieurstrames sm trames =
    foldl (\sm trame -> sendTrame sm trame) sm trames

    


-- | Extraire les n premiers messages  de chacune des 3  boîtes : Reception, Envoi, Spams de chaque compte du SmartMail
-- Le résultat est un nouveau compte SmartMail dont les 3 boîtes de messageries contiennent uniquement les n premiers messages.
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9]
-- >>> length $ spams $ obtenirCompte "tato.ange@smail.ca"  s
-- 8
-- >>> length $ spams $ obtenirCompte "tato.ange@smail.ca" $ extrairenMessages 2 s
-- 2
extrairenMessages :: Int -> SmartMail -> SmartMail
extrairenMessages n sm = 
    ajoutComptes allCs Map.empty
    where
        allCs = map (\(_,CompteSmail p r e s pref contacts)-> CompteSmail p (take n r) (take n e) (take n s) pref contacts) $ Map.toList sm

-- | Supprimer tous les messages (Reception, Envoi, Spam) du système Smartmail datant d'une date antérieure (strictement) à celle spécifiée
--
-- >>> s1 = ajoutComptes [csmail0, csmail1, csmail2, csmail3, csmail4, csmail5] emptySmartMail
-- >>> s = envoyerMessage_Plusieurstrames s1 [trame1, trame2,trame2,trame4,trame5,trame6,trame7,trame8,trame9, trame10, trame11, trame12]
-- >>> s3 = supprimerOldMessages (Date 2019 10 01) s
-- >>> nbTotalSpams s
-- 10
-- >>> length $ spams $ obtenirCompte "tato.ange@smail.ca" s3
-- 8
-- >>> s4 = supprimerOldMessages (Date 2020 01 01) s
-- >>> length $ spams $ obtenirCompte "tato.ange@smail.ca"  s4
-- 5
supprimerOldMessages :: Date -> SmartMail-> SmartMail
supprimerOldMessages date sm =
    ajoutComptes allCs Map.empty
    where
        allCs = map (\(_,CompteSmail p r e s pref contacts)-> 
            CompteSmail p (filterOld date r) (filterOld date e) (filterOld2 date s) pref contacts) $ Map.toList sm

filterOld d messages =
    filter (\m -> d >= date m) messages

filterOld2 d messages =
    filter (\(m,_) -> d >= date m) messages

-- | Reformater boîte
-- Afin de faciliter la gestion des messages, l'administrateur voudrait pouvoir reorganiser toutes les boîtes (reception, spams, envoi) de la manière suivante:
--       chaque boîte doit devenir un dictionnaire avec le courriel de l'emetteur comme clé  et comme valeur, une liste des messages reçus de cette personne. 
--       Les messages de cette liste doivent comporter uniquement la date (au format "jj/mm/aaaa"),
--       l'objet et le contenu du message (sous forme d'un triplet). 
-- Voudevez donc écrire la fonction reformaterBoite qui reformate une boîte selon le principe décrit.
-- Donnée: Une boîte de messages (une liste de trames)
-- Sortie: Le dictionnaire correspondant. 
--
-- >>> reformaterBoite [trame7, trame8, trame9, trame10, trame11]
-- fromList [("equipesmartmail@smail.ca",[("18/10/2020","un message","allo|allo"),("1/2/2020","un message","bien1venue"),("10/10/2020","un message","0de 3Bienvenue dans 1 et 9 | pour tp1et2"),("15/1/2020","un message","offre de Bienvenue dans  publicit\233 gratuit pour voyage special"),("10/18/2019","un message","Bienvenue dans sexe du viagra chaud nu")])]
reformaterBoite :: [Trame] -> Map.Map String [(String, String, String)]
reformaterBoite trames =
    foldl (
        \m (mail,triplet)->
            if (Map.member mail m) then
                Map.insert mail (triplet:( m Map.! mail)) $ Map.delete mail m
            else
                Map.insert mail [triplet] m 
        ) Map.empty formated
    where
        formated = map (\trame->(emetteur trame, (formatDate $ date trame, objet trame, contenu trame))) trames


formatDate (Date y m d) = show d ++ "/" ++ show m ++ "/" ++ show y

-- | Reformater compte
--  Utiliser la fonction reformaterBoite pour écrire la fonction reformaterCompte qui reconstruit  la structure d'un compte smail. Le résultat doit comporter l'ensemble des messages provenant des 
--  trois boîtes y compris celle des spams. Pour la boîte des spams, on ne s'interessera qu'aux Trames de message sans explications.
--  Donnée:  Un compte smail
--  Sortie:  Un triplet comportant les trois boîtes dans l'ordre et dans leur nouvelle format.
--
-- >>> ssm4 = Map.fromList [("nkambou.roger@smail.ca",CompteSmail (Personne "nkambou.roger@smail.ca" ("roger","nkambou")) [Trame (Entete (Date 2018 10 2) "Bienvenue" pers0 pers2) "Bienvenue dans votre boite smartMail !"] [Trame (Entete (Date 2018 10 26) "toto" pers2 pers1) "tato "] [] [] []),("tato.ange@smail.ca",CompteSmail (Personne "tato.ange@smail.ca" ("ange","tato")) [Trame (Entete (Date 2018 10 26) "toto" pers2 pers1) "tato ", Trame (Entete (Date 2018 10 25) "ggg" pers4 pers1) "jjjj",Trame (Entete (Date 2018 10 2) "Bienvenue" pers0 pers1) "Bienvenue dans votre boite smartMail !"] [] [(Trame (Entete (Date 2018 10 25) "ghgh jjhjh" pers3 pers1) "hgg6gg jhh7hh","Classique!! Contient 100% de mots comportant des caracteres etranges.")] [] []),("toto.tartampion@smail.ca",CompteSmail (Personne "toto.tartampion@smail.ca" ("tartampion","toto")) [] [Trame (Entete (Date 2018 10 25) "ghgh jjhjh" pers3 pers1) "hgg6gg jhh7hh",Trame (Entete (Date 2018 10 25) "ggg" pers3 pers1) "jjjj"] [] [] [])]
-- >>> reformaterCompte $ obtenirCompte "tato.ange@smail.ca" ssm4
-- (fromList [("equipesmartmail@smail.ca",[("2/10/2018","Bienvenue","Bienvenue dans votre boite smartMail !")]),("nkambou.roger@smail.ca",[("26/10/2018","toto","tato ")]),("noel.alice@smail.ca",[("25/10/2018","ggg","jjjj")])],fromList [],fromList [("robert.julien@smail.ca",[("25/10/2018","ghgh jjhjh","hgg6gg jhh7hh")])])
reformaterCompte :: CompteSmail
     -> (Map.Map String [(String, String, String)],
         Map.Map String [(String, String, String)],
         Map.Map String [(String, String, String)])
reformaterCompte cs =
    (
        reformaterBoite $ reception cs,
        reformaterBoite $ envoi cs,
        reformaterBoite $ map (\(a,_)->a) $ spams cs
        )


-- | Filtrage par prefrences
-- Filtre une trame recu selon les préférences du recepteur de ces trames.
-- Si la trame respecte toutes les préférences du recepteur alors retourner (NonSpam, Trame, "")
-- Si la trame ne respecte pas au moins une préférence du recepteur alors retourner (Spam, Trame, "probleme de preferences")
--
-- >>> s = ajoutCompte csmail6 emptySmartMail
-- >>> map (flip (filtragePreference) s) [trame14, trame15, trame16]
-- [(Spam,Trame (Entete (Date 2020 3 18) "Hola mimi" (Personne "pablo.adamo@blob.es" ("olivier","adam")) (Personne "mimi.lafleur@smail.ca" ("mimi","lafleur"))) "como estas ?","probleme de preferences"),(NonSpam,Trame (Entete (Date 2020 2 9) "Bingo" (Personne "ariane.carotte@techno.co" ("arianne","carotte")) (Personne "mimi.lafleur@smail.ca" ("mimi","lafleur"))) "J'ai trouve ce que tu cherchais hier",""),(NonSpam,Trame (Entete (Date 2020 1 7) "Par rapport a Ivan" (Personne "michel.desrosiers@blob.ca" ("michel","desrosiers")) (Personne "mimi.lafleur@smail.ca" ("mimi","lafleur"))) "Ivan ne viendra pas demain ?","")]
filtragePreference :: Trame -> SmartMail-> (TypeMessage, Trame, Explications)
filtragePreference trame sm =
    if decision then
        (NonSpam,trame,"")
    else
        (Spam,trame,"probleme de preferences")
    where
        receiver = receveur trame
        rd = sm Map.! receiver
        prefs = preferences rd
        decision = foldr (&&) True $ map (\pref -> pref trame) prefs
