module State
where

import Data.Array
import Data.Word

data Store = Store (Array Integer MWord) Integer Integer deriving (Show, Eq)
type Location = Integer
type MemorySize = Integer
type MWord = Word8

--  Checks if the expression holds, then passes d through, 
-- otherwise aborts execution with an error message.
assert :: Show a=>Bool->a->a
assert expr d = if expr then d else error $ " Assertion failed for "++(show d)

-- | Zero initialised store of a given size
newStore :: MemorySize -> Store
newStore size =
    Store words 0 (size - 1)
    where 
        words = array (0,size - 1) [(i,0) | i <- [0..(size - 1)]]

-- | Copies the contents of a data structure within the store,
-- copies MemorySize words from source location to target location. 
copy :: Location -> MemorySize -> Location -> Store -> Store
copy from n to store =
    insert to (extract from n store) store

-- | Get the top of the stack (lowest address in use)
getTop :: Store -> Location
getTop (Store _ heap stack) = stack

-- | Set the top of the stack (lowest address in use) 
setTop :: Location -> Store -> Store
setTop location (Store words heap stack) = Store words location stack

-- | Copies the list of words into the target location onwards, 
-- overwriting existing data
insert :: Location -> [MWord] -> Store -> Store
insert location insertWords (Store words heap stack) =
    Store newWords heap stack
    where
        -- 0..location - 1 from words
        -- location..location + length (insertWords) - 1 is insertWors
        -- location + length(insertWords)..size-1 from words
        newWords = listArray (toInteger 0, toInteger ((length resultingList) - 1)) resultingList
        elements = elems words
        resultingList = (take (fromIntegral location) elements) ++ insertWords ++ (drop ((fromIntegral location) + (length insertWords)) elements)
        (_, last) = bounds words

-- | Copies memory size numbers of words from source location to the list.
extract :: Location -> MemorySize -> Store -> [MWord]
extract location size (Store words heap stack) =
    take (fromIntegral size) $ drop (fromIntegral location) $ elems words

-- | Arithmetic on memory size and locations
addSize :: MemorySize -> MemorySize -> MemorySize 
addSize a b = a + b

subSize :: MemorySize -> MemorySize -> MemorySize 
subSize a b = a - b

multSize :: Integer -> MemorySize -> MemorySize 
multSize n s = n * s

addLocation :: Location -> MemorySize -> Location
addLocation l s = l + s

subLocation :: Location -> MemorySize -> Location
subLocation l s = l - s

-- | Concatenation property of store model 
concatInStoreAxiom str1 str2 loc store =
    assert ( str1 ++str2 == extract loc size store'') "Success"
    where
        size = toInteger $ length str1 + length str2 
        loc2 = addLocation loc ( toInteger $ length str1 ) 
        store' = insert loc str1 store
        store'' = insert loc2 str2 store'

-- unit tests

testNewStore = assert (actual == expected) "Success"
    where
        actual = newStore 10
        expected = Store (listArray (0,9) [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])  0 9
