import Control.Concurrent

data Board = Board Int [[Bool]] deriving Eq 
data Automaton = Automaton Int Int Int Int

instance Show Automaton where
    show (Automaton m1 n1 m2 n2) = "s " ++ (show m1) ++ " " ++ (show n1) ++ "\nb " ++ (show m2) ++ " " ++ (show n2) ++ "\n"

instance Show Board where
    show (Board n rows) =
        h ++ "\n" ++ (foldl (++) "" $ map showRow zippedRows)
        where
            h = header n
            zippedRows = zip (take n $ from 1) rows

header :: Int -> String
header n = "  " ++ (foldl (++) "" $ map (\x -> " " ++ (show $ rem x 10)) $ take n $ from 1)

showRow :: (Int, [Bool]) -> String
showRow (i, row) = (show i) ++ (if i<10 then " " else "") ++ (foldl (++) "" $ map (\x -> if x then " O" else " .") row) ++ "\n"

from :: Int -> [Int]
from n = n : (from $ n + 1)

create :: Int -> Board
create n = Board n  $ take n $ repeat $ take n $ repeat False

placeCells :: [(Int, Int)] -> Board -> Board
placeCells cells board = foldl (\b c->placeCell c b) board cells

placeCell :: (Int, Int) -> Board -> Board
placeCell (x,y) board = set board (x,y) True

makeEmpty :: [(Int, Int)] -> Board -> Board
makeEmpty cells board = foldl (\b c->makeEmpty1 c b) board cells

makeEmpty1 :: (Int, Int) -> Board -> Board
makeEmpty1 (x,y) board = set board (x,y) False

set :: Board -> (Int, Int) -> Bool -> Board
set (Board n rows) (x,y) value = Board n rows'
    where
        rows' = set' rows (x,y) value
        set' (row:rows) (x,1) value = (setValue row x value) : rows
        set' (row:rows) (x,y) value = row : (set' rows (x,y-1) value)
        setValue (e:row) 1 value = value:row
        setValue (e:row) x value = e:(setValue row (x - 1) value) 

redefineB :: Int -> Int -> Automaton -> Automaton
redefineB m2 n2 (Automaton m1 n1 _ _) = Automaton m1 n1 m2 n2

redefineS :: Int -> Int -> Automaton -> Automaton
redefineS m1 n1 (Automaton _ _ m2 n2) = Automaton m1 n1 m2 n2

getAllCells :: Board -> [(Int, Int)]
getAllCells board = 
    map (\(_,pos) -> pos) $ 
    filter (\(value,_)->value) $ getPositionMap board

getPositionMap :: Board -> [(Bool,(Int, Int))]
getPositionMap (Board n rows) =  
    map ((\(v,(x,y))->(v,(y,x)))) $
    foldl (++) [] $ 
    map (\(row, i) -> zip row $ map (\j->(i,j)) $ take n $ from 1 ) 
    $ zip rows $ take n $ from 1

nextGen :: Automaton -> Board -> Board
nextGen (Automaton m1 n1 m2 n2) board = 
    placeCells born $ placeCells survive $ create n
    where
        Board n _ = board
        all = getPositionMap $ board
        cells = map (\(_,pos) -> pos) $ filter (\(value,_)->value) all
        empty = map (\(_,pos) -> pos) $ filter (\(value,_)->not value) all
        cellsNeighbors = map (countNeighbors board) cells
        emptyNeighbors = map (countNeighbors board) empty
        survive = map (\(_,pos) -> pos) $ filter (\(n, _) -> m1 <= n && n <= n1) $ zip cellsNeighbors cells 
        born = map (\(_,pos) -> pos) $ filter (\(n, _) -> m2 <= n && n <= n2) $ zip emptyNeighbors empty 

countNeighbors :: Board -> (Int, Int) -> Int
countNeighbors board (x,y) = 
    isCell board (x - 1, y - 1) +
    isCell board (x, y - 1) +
    isCell board (x + 1, y - 1) +
    isCell board (x - 1, y) +
    isCell board (x + 1, y) +
    isCell board (x - 1, y + 1) +
    isCell board (x, y + 1) +
    isCell board (x + 1, y + 1)

isCell :: Board -> (Int, Int) -> Int
isCell (Board n rows) (x, y)
    | x <= 0 = 0
    | y <= 0 = 0
    | x > n = 0
    | y > n = 0
    | otherwise = isCell' rows (x,y) 
    where
        isCell' (row:rows) (x,1) = isCell'' row x
        isCell' (_:rows) (x,y) = isCell' rows (x, y - 1)
        isCell'' (e:row) 1 = if e then 1 else 0
        isCell'' (_:row) y = isCell'' row (y - 1)

live :: Int -> Automaton -> Board -> [Board]
live 0 _ _ = []
live n automaton board = 
    if board == nextBoard then
        [board]
    else
        board : (live (n - 1) automaton nextBoard)
    where
        nextBoard = nextGen automaton board

test = "5 (s 2 3, b 3 3) 3 3 1 3 2 1 3 2 2 3"

automatonBoardFromString :: String -> (Automaton, Board)
automatonBoardFromString s =
    (automaton, board)
    where
        (ns:l1:m1s:n1s:l2:m2s:n2s:more) = words s
        n = read ns :: Int
        bs1 = l1!!1
        n1 = read [n1s!!0] :: Int
        m1 = read m1s :: Int
        n2 = read [n2s!!0] :: Int
        m2 = read m2s :: Int 
        automaton = if bs1=='s' then
                        Automaton m1 n1 m2 n2 
                    else
                        Automaton m2 n2 m1 n1 
        lst = map (\x->read x :: Int) more
        pairs = pairsFunction lst
        board = placeCells pairs $ create n

pairsFunction [] = []
pairsFunction (x:y:t) = (x,y):(pairsFunction t)

buildOutput board automaton showRules showPositions isStable =
    (show board) ++ rules ++ positions ++ stable
    where
        rules =
            if showRules then
                "Rules for the game: \n" ++ (show automaton)
            else
                ""
        positions = 
            if showPositions then
                "Positions of all living cells: " ++(show $ getAllCells board) ++ "\n"
            else
                ""
        stable =
            if isStable then
                "A stable configuration is reached.\n"
            else
                ""

main :: IO ()
main = 
    loop (automaton, board, False, False, False)
    where
        (automaton, board) = automatonBoardFromString "1 (s 0 0, b 0 0)"

loop (automaton, board, showRules, showPositions, stable) =
    do
        line <- getLine
        if line == "quit" then
            return ()
        else
            do
                let (readFormFile, liveCounter, automaton', board', showRules', showPositions', stable') = processLine line (automaton, board, showRules, showPositions, stable)
                if liveCounter > 0 then
                    do
                        board'' <- doLive automaton' board' liveCounter
                        loop (automaton', board', showRules', showPositions', stable')
                else
                    if readFormFile /= "" then
                        do
                            content <- readFile readFormFile
                            let (automatonFromFile, boardFromFile) = automatonBoardFromString content
                            putStr (buildOutput boardFromFile automatonFromFile showRules' showPositions' stable')
                            loop (automatonFromFile, boardFromFile, showRules', showPositions', stable')
                    else
                        do
                            putStr (buildOutput board' automaton' showRules' showPositions' stable')
                            loop (automaton', board', showRules', showPositions', stable')

doLive automaton board n =
    if n == 0 then
        -- putStr ""
        return board
    else
        do
            let board' = nextGen automaton board
            if board' == board then
                do
                    putStr "A stable configuration is reached.\n"
                    return board
            else
                do
                    putStr (show board')
                    Control.Concurrent.threadDelay 500000
                    doLive automaton board' (n - 1)

processLine line (automaton, board, showRules, showPositions, stable) =
    r
    where
        parts = words line
        r = processCase parts
        processCase ["c", ns] = ("", 0, automaton, create (read ns :: Int), showRules, showPositions, False)
        processCase ("n" : cellss) = ("", 0, automaton, placeCells (pairsFunction (map (\x->read x :: Int) cellss)) board, showRules, showPositions, False)
        processCase ("e" : cellss) = ("", 0, automaton, makeEmpty (pairsFunction (map (\x->read x :: Int) cellss)) board, showRules, showPositions, False)
        processCase ["b", ms, ns] = ("", 0, redefineB (read ms :: Int) (read ns :: Int) automaton, board, showRules, showPositions, False)
        processCase ["s", ms, ns] = ("", 0, redefineS (read ms :: Int) (read ns :: Int) automaton, board, showRules, showPositions, False)
        processCase [] = ("", 0, automaton, board', showRules, showPositions, board' == board)
        processCase ["?"] = ("", 0, automaton, board, not showRules, showPositions, stable)
        processCase ["w"] = ("", 0, automaton, board, showRules, not showPositions, stable)
        processCase ["l", ns] = ("", read ns :: Int, automaton, board, showRules, showPositions, False)
        processCase ["r", filename] = (filename, 0, automaton, board, showRules, showPositions, stable)
        processCase _ = ("", 0, automaton, board, showRules, showPositions, stable)
        board' = nextGen automaton board
